#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <stdarg.h>
#include <netdb.h>
#include <errno.h>
#include <limits.h>
#include <signal.h>


extern int tcp_connect(const char *hstr, unsigned short port);

#define COIL_ON 0x00ff
#define COIL_OFF 0x0000
#define COIL_RELEASE 0xffff /* "release the force", whatever that is... */


#define CHECK_DI 2
#define FORCE_DO 5
static struct _CODE_ {
	int code;
	char *name;
} modbus_function[] = {
	{ 0x01, "Read coil status / Read Discrete Output bit" },
	{ CHECK_DI, "Read input status / Read Discrete Input bit" },
	{ 0x03, "Read holding registers / Read int or float process data" },
	{ 0x04, "Read input registers / Read int or float process data" },
	{ FORCE_DO, "Force single coil / Write data to force coil ON or OFF" },
	{ 0x06, "Preset single register / Write data in 16-bit int format" },
	{ 0x08, "Loopback diagnostics / Data-ping" },
	{ 0x0f, "Force multiple coils / Force multiple coil(s) ON or OFF" },
	{ 0x10, "Preset multiple registers / Preset multiple registers" },
};

#define ARRAY_SIZE(x) (sizeof(x)/sizeof(x[0]))
int debug = 0;

int timeout=60;
/* handles socket timeouts */
void
	socket_timeout_alarm_handler (int sig)
{
	if (sig == SIGALRM)
		printf ("CRITICAL - Socket timeout after %d seconds\n", timeout);
	else
		printf ("CRITICAL - Abnormal timeout after %d seconds\n", timeout);

	exit (3);
}

const char *get_func_name(int code)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(modbus_function); i++)
		if (modbus_function[i].code == code)
			return modbus_function[i].name;

	return "Unknown function";
}

static void usage(void)
{
	printf("Usage: check_adam -H <host> [-p <port>] -c <coil>\n");
	printf("(followed by EITHER)\n");
	printf("         --check-di --expect=[0|1]\n");
	printf("(OR)\n");
	printf("         --set-do --state=[on|off|release]\n");
	putchar('\n');
	printf("-H : --host          the address to connect to\n");
	printf("-p : --port          the port to connect to\n");
	printf("-c : --coil          which coil to set/query\n");
	printf("-h : --help          this cruft\n");
	printf("-d : --debug         increase verbosity\n");
	printf("-t : --timeout       timeout in seconds (default %d)\n", timeout);
	printf("   : --check-di      check status of a digital input channel\n");
	printf("   : --expect=[0|1]  what to expect from a --check-di request\n");
	printf("   : --set-do        set status of a digital output channel\n");
	printf("   : --state=[on|off|release]  status to set the digital output channel to\n");
	putchar('\n');
}

void die(const char *fmt, ...)
{
	va_list ap;

	printf("fatal: ");
	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);

	exit(3);
}


/* we cheat a little. unit_id, func_code and data_len are actually in the
 * body, but they're always present, so we pretend they're in the head */
struct adam_header {
	unsigned short trans_id;
	unsigned short protocol;
	unsigned short body_len;
	unsigned char unit_id;
	unsigned char func_code;
};

struct func2_body {
	unsigned short start_addr;
	unsigned short req_reg;
};

struct func2_command {
	struct adam_header head;
	struct func2_body body;
};

#define ON 0x00ff
#define OFF 0
#define RELEASE 0xffff
struct func5_body {
	unsigned short coil_addr;
	unsigned short coil_state;
};

struct func5_command {
	struct adam_header head;
	struct func5_body body;
};

void print_header(struct adam_header *head)
{
	if (!debug)
		return;
	printf("trans_id:  %d\n", head->trans_id);
	printf("protocol:  %d\n", head->protocol);
	printf("body_len:  %d\n", htons(head->body_len));
	printf("unit_id:   %d\n", head->unit_id);
	printf("func_code: %d\n", head->func_code);
	puts("##########################################");
}

static inline int send_command(int sock, const void *buf, size_t len)
{
	return write(sock, buf, len);
}

static inline int
read_response(int sock, struct adam_header *head, void *buf, size_t buflen)
{
	int len;

	len = read(sock, head, sizeof(struct adam_header));

	/* since we cheat with the struct adam_header, we also have to
	 * recalculate the body length */
	len += read(sock, buf, ntohs(head->body_len) - 2);

	if (debug)
		printf("read %d bytes of response in total\n", len);

	return len;
}

static void bprint(const char *buf, int len)
{
	int i;

	if (!debug)
		return;

	if (len < 0 || !buf)
		return;

	for (i = 0; i < len; i++) {
		char c = buf[i];
		int b;

		for (b = CHAR_BIT; b; b--) {
			if (b == 4)
				putchar(' ');
			if ((c >> b) & 1)
				putchar('1');
			else
				putchar('0');
		}
		printf("  ");
	}

	putchar('\n');
}

static int check_DI_state(int sock, int coil)
{
	struct func2_command cmd;
	struct adam_header head;
	unsigned short regs;
	char buf[512];
	int len, i;

	memset(&cmd, 0, sizeof(cmd));
	cmd.head.trans_id = htons(1);
	cmd.head.protocol = 0;
	cmd.head.body_len = htons(6);
	cmd.head.unit_id = htons(1);
	cmd.head.func_code = 2;
	cmd.body.start_addr = htons(0);
	cmd.body.req_reg = htons(12);

	print_header(&cmd.head);
	send_command(sock, &cmd, sizeof(cmd));
	len = read_response(sock, &head, &buf, sizeof(buf));
	print_header(&head);

	bprint(buf, len - sizeof(struct adam_header));
	memcpy(&regs, buf + 1, sizeof(regs));

	if (debug) {
		printf("regs = %d\n", regs);
		bprint((const char *)&regs, 2);
		printf("0x%4x\n", regs);
		for (i = 0; i < 12; i++) {
			printf("DI %d is %s\n", i, (1 << i) & regs ? "not set" : "a closed circuit");
		}
	}

	return !!((1 << coil) & regs);
}

#define DO_BASE_ADDRESS 16
static int set_DO_state(int sock, int coil, unsigned short state)
{
	struct func5_command cmd, resp;
	struct adam_header head;
	char buf[512];

	coil += DO_BASE_ADDRESS;
	printf("Setting state of digital output at address %d to ", coil);
	switch (state) {
	case COIL_ON:
		puts("on");
		break;
	case COIL_OFF:
		puts("off");
		break;
	case COIL_RELEASE:
		puts("release");
		break;
	default:
		puts("something unknown... Nah. Returning instead\n");
		return 1;
		break;
	}
	memset(&cmd, 0, sizeof(cmd));
	cmd.head.trans_id = htons(1);
	cmd.head.protocol = 0;
	cmd.head.body_len = htons(6);
	cmd.head.unit_id = htons(1);
	cmd.head.func_code = 5;
	cmd.body.coil_addr = htons(coil);
	cmd.body.coil_state = state;

	send_command(sock, &cmd, sizeof(cmd));
	read_response(sock, &head, &buf, sizeof(buf));
	print_header(&head);
	bprint(buf, ntohs(head.body_len));
	memcpy(&resp, &head, sizeof(head));
	memcpy(&resp + sizeof(head), buf, sizeof(resp) - sizeof(head));

	return memcmp(&resp, &cmd, sizeof(cmd));
}

int main(int argc, char **argv)
{
	int i, retval = 0, sock, func = 0, state = COIL_ON;
	const char *host = NULL;
	int port = 0, coil = -1;
	int expect = 1;

	for (i = 1; i < argc; i++) {
		char *arg = argv[i];
		char *opt = NULL;

		if (*arg != '-' && !host) {
			host = arg;
			continue;
		}

		opt = strchr(arg, '=');
		if (opt) {
			*opt = 0;
			opt++;
		}
		if (debug)
			printf("arg = '%s'; opt = '%s'\n", arg, opt);

		/* arguments *without* options first */
		if (!strcmp(arg, "--help") || !strcmp(arg, "-h")) {
			usage();
			exit(0);
		}
		if (!strcmp(arg, "--check-di")) {
			func = CHECK_DI;
			continue;
		}
		if (!strcmp(arg, "--set-do")) {
			func = FORCE_DO;
			continue;
		}
		if (!strcmp(arg, "--debug") || !strcmp(arg, "-d")) {
			debug++;
			continue;
		}

		/* arguments *with* options below */
		if (!opt && i < argc - 1)
			opt = argv[++i];

		if (!opt)
			die("Argument %s requires a parameter.\n", arg);

		if (!strcmp(arg, "--host") || !strcmp(arg, "-H")) {
			host = opt;
			continue;
		}
		if (!strcmp(arg, "--port") || !strcmp(arg, "-p")) {
			port = strtoul(opt, NULL, 0);
			if (!port || port > 65535)
				die("Port must be between 1 and 65535. %s is not good enough\n", opt);
			continue;
		}
		if (!strcmp(arg, "--coil") || !strcmp(arg, "-c")) {
			coil = strtoul(opt, NULL, 0);
			continue;
		}
		if (!strcmp(arg, "--timeout") || !strcmp(arg, "-t")) {
			timeout = strtoul(opt, NULL, 0);
			continue;
		}
		if (!strcmp(arg, "--expect")) {
			expect = *opt - '0';
			if (expect < 0 || expect > 1)
				die("--expect takes only 1 or 0 as values. %s is no good\n", opt);
			continue;
		}
		if (!strcmp(arg, "--state")) {
			if (!strcmp(opt, "on"))
				state = COIL_ON;
			else if (!strcmp(opt, "off"))
				state = COIL_OFF;
			else if (!strcmp(opt, "release"))
				state = COIL_RELEASE;
			else
				die("--state requires an option of either 'on', 'off' or 'release'\n");
			continue;
		}
		die("Unrecognized argument '%s'\n", arg);
	}

	/* argument sanity checking */
	if (!port)
		port = 502; /* modbus default port */

	if (!host)
		die("No IP-address or hostname specified\n");

	if (!func)
		die("You must specify either '--check-di' or '--set-do'\n");

	if (coil == -1)
		die("--coil argument not specified\n");

	switch (func) {
	case CHECK_DI:
		if (coil < 0 || coil > 11)
			die("Register to check must be between 0 and 11, inclusive.\n");
		break;
	case FORCE_DO:
		if (coil < 0 || coil > 1)
			die("DO coil to set must be 0 or 1.\n");
		break;
	default:
		die("Either --set-do or --check-di must be specified\n");
		break;
	}

	/* set up the timer */
	signal (SIGALRM, socket_timeout_alarm_handler);
	alarm (timeout);
	/* things look ok, so go ahead and try to connect */
	sock = tcp_connect(host, port);

	switch (func) {
	case CHECK_DI:
		retval = check_DI_state(sock, coil);
		if (debug)
			printf("retval: %d, expect: %d\n", retval, expect);
		if (retval != expect)
			printf("CRITICAL: ");
		printf("DI Coil %d is%s a closed circuit\n", coil, retval ? " not" : "");
		if (retval != expect)
			retval = 2;
		else
			retval = 0;
		break;

	case FORCE_DO:
		retval = set_DO_state(sock, coil, state);
		break;
	}

	if (debug)
		printf("retval = %d\n", retval);

	close(sock);

	return retval;
}
