#include <stdio.h>
#include <fcntl.h>
#include <netdb.h>
#include <unistd.h>

#include <sys/poll.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <string.h>
#include <errno.h>

#include <time.h>

extern void die(const char *fmt, ...);

/* checks for data to read on the socket
 * returns positive if there's data inbound, and 0 if not
 * handles its own error checking */
static int my_poll(int sock, int timeout, unsigned short flags)
{
	struct pollfd pfd;
	int retval;

	pfd.fd = sock;
	pfd.events = flags;

	retval = poll(&pfd, 1, timeout);
	if (!retval)
		return 0;

	if (retval < 0)
		die("poll() failed: %s\n", strerror(errno));

	return pfd.revents;
}

/* gets a socket and connects to port 'port' on host 'hstr' if
 * the connection can be established within 'timeout' milliseconds.
 * Returns a socket identifier on success, and -1 on failure */
int tcp_connect(const char *hstr, unsigned short port)
{
	struct sockaddr_in sain;
	struct sockaddr *sa = (struct sockaddr *)&sain;
	struct hostent *host;
	int sock, retval;

	memset(&sain, 0, sizeof(sain));

	host = gethostbyname(hstr);
	if (!host)
		die("Unable to look up host '%s': %s\n", hstr);

	/* make sure we're on the net */
	if(host->h_addrtype != AF_INET)
		die("bad broadcast address");

	memset(&sain, 0, sizeof(sain));
	sain.sin_family = host->h_addrtype;
	memcpy(&sain.sin_addr, host->h_addr_list[0], host->h_length);
	sain.sin_port = htons(port);

	sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (sock == -1)
		die("Failed to create socket: %s\n", strerror(errno));

	retval = connect(sock, sa, sizeof(sain));
	if (retval < 0)
		die("connect() failed: %s\n", strerror(errno));

	return sock;
}

/* Sends data through connected and properly set up socket
 * Returns the number of bytes successfully sent
 */
int tcp_send(int sock, void *data, int len, int timeout)
{
	int left, resp, written = 0;

	left = len;

	while(left > 0) {

		if(!(my_poll(sock, timeout, POLLOUT))) return written;
		
		resp = write(sock, data, left);
		
		if(resp > 0) {
			data = ((char *)data) + resp;
			written += resp;
			left -= resp;
		} else if(errno != EINTR) return written;
	}

	return len;
}

/* Reads data from socket and stores in *data
 * Returns number of bytes read */
int tcp_receive(int sock, void *data, int len, int timeout)
{
	int left, resp, rcvd = 0;

	left = len;

	while(left > 0) {
		if(!(my_poll(sock, timeout, POLLIN | POLLPRI)))
			return (rcvd);

		resp = read(sock, data, left);
		if(resp > 0) {
			data = ((char *)data) + resp;
			rcvd += resp;
			left -= resp;
		}
		else if(errno != EINTR)
			return rcvd;
	}

	return len;
}
